"""

    Account class for convex api


"""
import secrets

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey

from eth_utils import (
    remove_0x_prefix,
    to_bytes,
    to_hex
)

from mnemonic import Mnemonic

from convex_api.utils import (
    to_address,
    to_public_key_checksum
)


class Account:

    def __init__(self, private_key, address=None, name=None):
        """

        Create a new account with a private key as a Ed25519PrivateKey. It is better to use
        one of the following static methods to create an Account object:

            * :meth:`.create`
            * :meth:`import_from_bytes`
            * :meth:`import_from_file`
            * :meth:`import_from_mnemonic`
            * :meth:`import_from_text`

        :param Ed25519PrivateKey private_key: The public/private key as an Ed25519PrivateKey object

        :param int address: address of the account


        The Convex account class, contains the public/private keys and possibly an address.

        You can create a new account object, it will only have it's public/private keys but does not have a valid account address.
        To obtain a new account address, you need to call the :py:meth:`.ConvexAPI.create_account` with the new account object.

        This is so that you can use the same public/private keys for multiple convex accounts.

        Once you have your new account you need to save the public/private keys using the `export..` methods and also you
        need to save the account address.

        To re-use the account again, you can import the keys and set the account address using one of the `import..` methods.

        **Note**
        For security reasons all of the keys, and text returned by the accounts are only truncated ending with a **`...`**

        .. code-block:: python

            >>> # import convex-api
            >>> from convex_api import ConvexAPI

            >>> # setup the network connection
            >>> convex_api = ConvexAPI('https://convex.world')

            >>> # create a new account and address
            >>> account = convex_api.create_account()

            >>> # export the private key to a file
            >>> account.export_to_file('/tmp/my_account.pem', 'my secret password')

            >>> # save the address for later
            >>> my_address = account.address

            >>> # ----

            >>> # now import the account and address for later use
            >>> reload_account = Account.import_from_file('/tmp/my_account.pem', 'my secret password', my_address)

        """
        self._private_key = private_key
        self._public_key = private_key.public_key()
        self._address = None
        if address is not None:
            self._address = to_address(address)
        self._name = name

    def sign(self, hash_text):
        """

        Sign a hash text using the private key.

        :param str hash_text: Hex string of the hash to sign

        :returns: Hex string of the signed text

        .. code-block:: python

            >>> # create an account with no address
            >>> account = Account.create()
            >>> # sign a given hash
            >>> sig = account.sign('7e2f1062f5fc51ed65a28b5945b49425aa42df6b7e67107efec357794096e05e')
            >>> print(sig)
            '5d41b964c63d1087ad66e58f4f9d3fe2b7bd0560b..'

        """
        hash_data = to_bytes(hexstr=hash_text)
        signed_hash_bytes = self._private_key.sign(hash_data)
        return to_hex(signed_hash_bytes)

    def export_to_text(self, password):
        """

        Export the private key to an encrypted PEM string.

        :param str password: Password to encrypt the private key value

        :returns: The private key as a PEM formated encrypted string

        .. code-block:: python

            >>> # create an account with the network
            >>> account = convex_api.create_account()

            >>> # export the private key for later use
            >>> print(account.export_to_text('secret password'))
            -----BEGIN ENCRYPTED PRIVATE KEY-----
            MIGbMFcGCSqGSIb3DQEFDTBKMCkGCSqGSIb3DQEFDDAcBAhKG+LC3hJoJQICCAAw
            DAYIKoZIhvcNAgkFAD ...


        """
        if isinstance(password, str):
            password = password.encode()
        private_data = self._private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.BestAvailableEncryption(password)
        )
        return private_data.decode()

    @property
    def export_to_mnemonic(self):
        """

        Export the private key as a mnemonic words. You must keep this secret since the private key can be
        recreated using the words.

        :returns: mnemonic word list of the private key

        .. code-block:: python

            >>> # create an account with the network
            >>> account = convex_api.create_account()

            >>> # export the private key for later use
            >>> print(account.export_to_mnemonic())
            grief stuff resemble dry frozen exercise ...

        """
        mnemonic = Mnemonic('english')
        return mnemonic.to_mnemonic(self._private_key.private_bytes(
            encoding=serialization.Encoding.Raw,
            format=serialization.PrivateFormat.Raw,
            encryption_algorithm=serialization.NoEncryption()
        ))

    def export_to_file(self, filename, password):
        """

        Export the private key to a file. This uses `export_to_text` to export as a string.
        Then saves this in a file.

        :param str filename: Filename to create with the PEM string

        :param str password: Password to use to encypt the private key

        .. code-block:: python

            >>> # create an account with the network
            >>> account = convex_api.create_account()

            >>> # export the private key to a file
            >>> account.export_to_file('my_account.pem', 'secret password')


        """
        with open(filename, 'w') as fp:
            fp.write(self.export_to_text(password))

    def __str__(self):
        return f'Account {self.address}'

    @property
    def is_address(self):
        """

        Return true if the address for this account object is set

        :returns: True if this object has a valid address

        """
        return self._address is not None

    @property
    def address(self):
        """

        :returns: the network account address
        :rtype: int

        .. code-block:: python

            >>> # create an account with the network
            >>> account = convex_api.create_account()
            >>> print(account.address)
            42

        """
        return self._address

    @address.setter
    def address(self, value):
        """

        Sets the network address of this account

        :param value: Address to use for this account
        :type value: str, int

        .. code-block:: python

            >>> # import the account keys
            >>> account = Account.import_from_mnemonic('my private key words ..')

            >>> # set the address that was given to us when we created the account on the network
            >>> account.address = 42

        """
        self._address = to_address(value)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def public_key_bytes(self):
        """

        Return the public key address of the account in the byte format

        :returns: Address in bytes
        :rtype: byte

        .. code-block:: python

            >>> # create an account with the network
            >>> account = convex_api.create_account()

            >>> # show the public key as bytes
            >>> print(account.public_key_bytes)
            b'6\\xd8\\xc5\\xc4\\r\\xbe-\\x1b\\x011\\xac\\xf4\\x1c8..

        """
        public_key_bytes = self._public_key.public_bytes(
            encoding=serialization.Encoding.Raw,
            format=serialization.PublicFormat.Raw
        )
        return public_key_bytes

    @property
    def public_key(self):
        """

        Return the public key of the account in the format '0x....'

        :returns: public_key with leading '0x'
        :rtype: str

        .. code-block:: python

            >>> # create an account with the network
            >>> account = convex_api.create_account()

            >>> # show the public key as a hex string
            >>> print(account.public_key)
            0x36d8c5c40dbe2d1b0131acf41c38b9d37ebe04d85...

        """
        return to_hex(self.public_key_bytes)

    @property
    def public_key_api(self):
        """

        Return the public key of the account without the leading '0x'

        :returns: public_key without the leading '0x'
        :rtype: str

        .. code-block:: python

            >>> # create an account with the network
            >>> account = convex_api.create_account()

            >>> # show the public key as a hex string with the leading '0x' removed
            >>> print(account.public_key_api)
            36d8c5c40dbe2d1b0131acf41c38b9d37ebe04d85...


        """
        return remove_0x_prefix(self.public_key)

    @property
    def public_key_checksum(self):
        """

        Return the public key of the account with checksum upper/lower case characters

        :returns: str public_key in checksum format

        .. code-block:: python

            >>> # create an account with the network
            >>> account = convex_api.create_account()

            >>> # show the public key as a hex string in checksum format
            >>> print(account.public_key_checksum)
            0x36D8c5C40dbE2D1b0131ACf41c38b9D37eBe04D85...

        """

        return to_public_key_checksum(self.public_key)

    @staticmethod
    def create():
        """

        Create a new account with a random key and an empty address.

        :returns: New Account object
        :rtype: Account

        .. code-block:: python

            >>> # create an account with no address
            >>> account = Account.create()
            >> account.is_address
            False

            >>> # create two accounts with the same public/private keys
            >>> account_1 = convex_api.create_account(account)
            >>> print(account_1.address)
            42

            >>> account_2 = convex_api.create_account(account)
            >>> print(account_2.address)
            43



        """
        return Account(Ed25519PrivateKey.generate())

    @staticmethod
    def import_from_bytes(value, address=None, name=None):
        """

        Import an account from a private key in bytes.

        :param address: address of the account
        :type address: str, int, optional

        :returns: Account object with the private/public key
        :rtype: Account

        .. code-block:: python

            >>> # create an account object from a raw private key
            >>> account = Account.import_from_bytes(0x0x973f69bcd654b26475917072...)


        """
        return Account(Ed25519PrivateKey.from_private_bytes(value), address=address, name=name)

    @staticmethod
    def import_from_text(text, password, address=None, name=None):
        """

        Import an accout from an encrypted PEM string.

        :param str text: PAM text string with the encrypted key text

        :param str password: password to decrypt the private key

        :param address: address of the account
        :type address: str, int, optional

        :returns: Account object with the public/private key
        :rtype: Account

        .. code-block:: python

            >>> # create an account object from a enrcypted pem text
            >>> pem_text = '''-----BEGIN ENCRYPTED PRIVATE KEY-----
                MIGbMFcGCSqGSIb3DQEFDTBKMCkGCSqGSIb3DQEFDDAcBAi3qm1zgjCO5gICCAAw
                DAYIKoZIhvcNAgkFADAdBglghkgBZQMEASoEENjvj1n...
            '''
            >>> account = Account.import_from_text(pem_text, 'my secret password')


        """
        if isinstance(password, str):
            password = password.encode()
        if isinstance(text, str):
            text = text.encode()

        private_key = serialization.load_pem_private_key(text, password, backend=default_backend())
        if private_key:
            return Account(private_key, address=address, name=name)

    @staticmethod
    def import_from_mnemonic(words, address=None, name=None):
        """

        Creates a new account object using a list of words. These words contain the private key and must be kept secret.

        :param str words: List of mnemonic words to read

        :param address: address of the account
        :type address: str, int, optional

        :returns: Account object with the public/private key
        :rtype: Account

        .. code-block:: python

            >>> # create an account object from a mnemonic word list with the account address 42
            >>> # the address in this example has been created by using the method `ConvexAPI.create_account`

            >>> account = Account.import_from_text('my word list that is the private key ..', 42)

        """

        mnemonic = Mnemonic('english')
        value = mnemonic.to_entropy(words)
        return Account(Ed25519PrivateKey.from_private_bytes(value), address=address, name=name)

    @staticmethod
    def import_from_file(filename, password, address=None, name=None):
        """

        Load the encrypted private key from file. The file is saved in PEM format encrypted with a password

        :param str filename: Filename to read

        :param str password: password to decrypt the private key

        :param address: address of the account
        :type address: str, int, optional

        :returns: Account with the private/public key
        :rtype: Account

        .. code-block:: python

            >>> # create an account object from a enrcypted pem saved in a file
            >>> account = Account.import_from_file(my_account_key.pem, 'my secret password')


        """
        with open(filename, 'r') as fp:
            return Account.import_from_text(fp.read(), password, address=address, name=name)

    @staticmethod
    def import_from_account(account, address=None, name=None):
        password = secrets.token_hex(64)
        text = account.export_to_text(password)
        return Account.import_from_text(text, password, address=address, name=name)



"""

    Conftest.py

"""

import logging
import pytest
from eth_utils import to_bytes

from convex_api.account import Account
from convex_api.convex_api import ConvexAPI

logging.basicConfig(level=logging.DEBUG)
logging.getLogger('urllib3').setLevel(logging.INFO)

PRIVATE_TEST_KEY = 0x973f69bcd654b264759170724e1e30ccd2e75fc46b7993fd24ce755f0a8c24d0
PUBLIC_KEY = '0x5288fec4153b702430771dfac8aed0b21cafca4344dae0d47b97f0bf532b3306'


PRIVATE_TEST_KEY_TEXT = """
-----BEGIN ENCRYPTED PRIVATE KEY-----
MIGbMFcGCSqGSIb3DQEFDTBKMCkGCSqGSIb3DQEFDDAcBAi3qm1zgjCO5gICCAAw
DAYIKoZIhvcNAgkFADAdBglghkgBZQMEASoEENjvj1nzc0Qy22L+Zi+n7yIEQMLW
o++Jzwlcg3PbW1Y2PxicdFHM3dBOgTWmGsvfZiLhSxTluXTNRCZ8ZLL5pi7JWtCl
JAr4iFzPLkM18YEP2ZE=
-----END ENCRYPTED PRIVATE KEY-----
"""
PRIVATE_TEST_KEY_PASSWORD = 'secret'

CONVEX_URL = 'https://convex.world'

TEST_ACCOUNT_NAME = 'test.convex-api'

@pytest.fixture(scope='module')
def test_account_info():
    return {
        'private_hex' : PRIVATE_TEST_KEY,
        'private_bytes': to_bytes(PRIVATE_TEST_KEY),
        'private_text': PRIVATE_TEST_KEY_TEXT,
        'private_password': PRIVATE_TEST_KEY_PASSWORD,
        'public_key': PUBLIC_KEY
    }

@pytest.fixture(scope='module')
def test_account(convex, test_account_info):
    import_account = Account.import_from_bytes(test_account_info['private_bytes'])
    account = convex.setup_account(TEST_ACCOUNT_NAME, import_account)
    convex.topup_account(account)
    return account

@pytest.fixture(scope='module')
def convex_url():
    return CONVEX_URL

@pytest.fixture(scope='module')
def convex(convex_url):
    api = ConvexAPI(convex_url)
    return api

@pytest.fixture(scope='module')
def other_account(convex):
    account = convex.create_account()
    convex.topup_account(account)
    return account



"""

    Convex API Python Library

"""

from convex_api.account import Account                   # noqa: F401
from convex_api.convex_api import ConvexAPI              # noqa: F401


__version__ = "__version__ = '0.1.1'"



"""

    Exceptions

"""


class ConvexBaseError(Exception):
    def __init__(self, source, code, text):
        self.source = source
        self.code = code
        self.text = text

    def __str__(self):
        return f'{self.source}: {self.code} {self.text}'


class ConvexAPIError(ConvexBaseError):
    """ Raised when a request error occurs """


class ConvexRequestError(ConvexBaseError):
    """ Raised when a error-code is returned from a web call """



"""


    Test  Utils

"""
import secrets

PUBLIC_KEY = '0x5288fec4153b702430771dfac8aed0b21cafca4344dae0d47b97f0bf532b3306'
PUBLIC_KEY_CHECKSUM = '0x5288Fec4153b702430771DFAC8AeD0B21CAFca4344daE0d47B97F0bf532b3306'

from convex_api.utils import (
    is_address,
    is_public_key,
    is_public_key_hex,
    is_public_key_checksum,
    to_address,
    to_public_key_checksum
)

def test_utils_is_public_key_hex():
    public_key = secrets.token_hex(32)
    assert(is_public_key_hex(public_key))

def test_utils_is_public_key():
    public_key = secrets.token_hex(32)
    assert(is_public_key(public_key))

def test_utils_is_public_key_checksum():
    public_key = secrets.token_hex(32)
    public_key_checksum = to_public_key_checksum(public_key)
    assert(is_public_key_checksum(public_key_checksum))

def test_utils_to_public_key_checksum():
    # generate a ethereum public_key
    # convex public_key to checksum
    public_key_checksum = to_public_key_checksum(PUBLIC_KEY)
    assert(is_public_key_checksum(public_key_checksum))
    assert(public_key_checksum == PUBLIC_KEY_CHECKSUM)

def test_utils_is_address():
    address_int = secrets.randbelow(pow(2, 1024)) + 1
    assert(is_address(address_int))

    address_str = str(address_int)
    assert(is_address(address_str))

    address = to_address(f'#{address_str}')
    assert(address == address_int)

    assert(not is_address('test'))
    assert(not is_address(' #'))
    assert(is_address('#0'))
    assert(not is_address('#-1'))



"""
    Test signing to send a transaction

"""
import json
import requests

from eth_utils import to_hex, remove_0x_prefix, to_bytes

from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey
from cryptography.hazmat.primitives import serialization

PRIVATE_TEST_KEY = 0x973f69bcd654b264759170724e1e30ccd2e75fc46b7993fd24ce755f0a8c24d0
PUBLIC_ADDRESS = '5288fec4153b702430771dfac8aed0b21cafca4344dae0d47b97f0bf532b3306'

def get_test_account():
    private_key = Ed25519PrivateKey.from_private_bytes(to_bytes(PRIVATE_TEST_KEY))

    public_key = private_key.public_key()
    public_key_bytes = public_key.public_bytes(
        encoding=serialization.Encoding.Raw,
        format=serialization.PublicFormat.Raw
    )
    public_address = remove_0x_prefix(to_hex(public_key_bytes))
    assert(public_address == PUBLIC_ADDRESS)
    return private_key, public_address


def create_account(convex_url, public_address):
    account_data = {
        'accountKey': public_address,
    }
    print('create_account send', account_data)
    url = f'{convex_url}/api/v1/createAccount'
    response = requests.post(url, data=json.dumps(account_data))
    assert(response.status_code == 200)
    result = response.json()
    return result['address']

def request_funds(convex_url, address):
    faucet_data = {
        'address': address,
        'amount': 10000000
    }
    print('faucet send', faucet_data)
    url = f'{convex_url}/api/v1/faucet'
    response = requests.post(url, data=json.dumps(faucet_data))
    if response.status_code != 200:
        print('error', response.text)
        assert(response.status_code == 200)
    print('faucet response', response.json())


def test_submit_transaction(convex_url):

    private_key, public_address = get_test_account()
    address = create_account(convex_url, public_address)
    request_funds(convex_url, address)
    # faucet request amount

    # prepare
    prepare_data = {
        'address': address,
        'source': '(map inc [1 2 3])'
    }
    url = f'{convex_url}/api/v1/transaction/prepare'
    print('prepare send', prepare_data)
    response = requests.post(url, data=json.dumps(prepare_data))
    if response.status_code != 200:
        print('prepare error', response.text)
        assert(response.status_code == 200)

    result = response.json()

    #submit
    print(result)
    hash_data = to_bytes(hexstr=result['hash'])
    signed_hash_bytes = private_key.sign(hash_data)
    signed_hash = to_hex(signed_hash_bytes)
    submit_data = {
        'address': address,
        'accountKey': public_address,
        'hash': result['hash'],
        'sig': remove_0x_prefix(signed_hash)
    }

    url = f'{convex_url}/api/v1/transaction/submit'
    print('submit send', submit_data)
    response = requests.post(url, data=json.dumps(submit_data))
    if response.status_code != 200:
        print('submit error', response.text, response.status_code)
        assert(response.status_code == 200)
    print(response.json())



def test_query_lisp(convex_url):
    private_key, public_address = get_test_account()
    address = create_account(convex_url, public_address)
    request_funds(convex_url, address)
    query_data = {
        'address': address,
        'lang': 'convex-lisp',
        'source': f'(balance {address})'
    }
    url = f'{convex_url}/api/v1/query'
    print('query send', query_data)
    response = requests.post(url, data=json.dumps(query_data))
    if response.status_code != 200:
        print('query error', response.text)
        assert(response.status_code == 200)
    result = response.json()
    assert(result)
    assert(result['value'] > 0)


def test_query_scrypt(convex_url):
    private_key, public_address = get_test_account()
    address = create_account(convex_url, public_address)
    request_funds(convex_url, address)
    query_data = {
        'address': address,
        'lang': 'convex-scrypt',
        'source': f'balance({address})'
    }
    url = f'{convex_url}/api/v1/query'
    print('query send', query_data)
    response = requests.post(url, data=json.dumps(query_data))
    if response.status_code != 200:
        print('query error', response.text)
        assert(response.status_code == 200)
    result = response.json()
    assert(result)
    print(result)
    assert(result['value'] > 0)




"""

    Test Convex api

"""
import pytest
import secrets
from eth_utils import (
    remove_0x_prefix,
    add_0x_prefix
)

from convex_api.account import Account
from convex_api.convex_api import ConvexAPI
from convex_api.exceptions import (
    ConvexAPIError,
    ConvexRequestError
)

from convex_api.utils import to_address

TEST_FUNDING_AMOUNT = 8888888


def test_convex_api_language_setup(convex_url):
    convex = ConvexAPI(convex_url)
    assert(convex.language == ConvexAPI.LANGUAGE_LISP)

    convex = ConvexAPI(convex_url, ConvexAPI.LANGUAGE_LISP)
    assert(convex.language == ConvexAPI.LANGUAGE_LISP)

    convex = ConvexAPI(convex_url, ConvexAPI.LANGUAGE_SCRYPT)
    assert(convex.language == ConvexAPI.LANGUAGE_SCRYPT)

    with pytest.raises(ValueError, match='language'):
        convex = ConvexAPI(convex_url, "bad-language")

def test_convex_api_request_funds(convex_url, test_account):
    convex = ConvexAPI(convex_url)
    amount = secrets.randbelow(100) + 1
    request_amount = convex.request_funds(amount, test_account)
    assert(request_amount == amount)

def test_convex_api_topup_account(convex_url):
    convex = ConvexAPI(convex_url)
    account = convex.create_account()
    topup_amount = TEST_FUNDING_AMOUNT
    amount = convex.topup_account(account, topup_amount)
    assert(amount >= topup_amount)

    account = convex.create_account()
    amount = convex.topup_account(account)
    assert(amount >= 0)

def test_convex_get_account_info(convex_url, test_account):
    convex = ConvexAPI(convex_url)
    info = convex.get_account_info(test_account)
    assert(info)
    assert(info['type']== 'user')
    assert(info['balance'] > 0)
    assert(info['sequence'] >= 0)

    with pytest.raises(ConvexRequestError, match='INCORRECT'):
        info = convex.get_account_info(pow(2, 100))

    with pytest.raises(ConvexRequestError, match='INCORRECT'):
        info = convex.get_account_info(pow(2, 1024))

    account = convex.create_account()
    request_amount = convex.request_funds(TEST_FUNDING_AMOUNT, account)
    info = convex.get_account_info(account)
    assert(info)
    assert(info['balance'] == TEST_FUNDING_AMOUNT)

def test_convex_api_send_basic_lisp(convex_url, test_account):
    convex = ConvexAPI(convex_url)
    request_amount = convex.request_funds(TEST_FUNDING_AMOUNT, test_account)
    result = convex.send('(map inc [1 2 3 4 5])', test_account)
    assert 'value' in result
    assert(result['value'] == [2, 3, 4, 5, 6])

def test_convex_api_send_basic_scrypt(convex_url, test_account):
    convex = ConvexAPI(convex_url, ConvexAPI.LANGUAGE_SCRYPT)
    request_amount = convex.request_funds(TEST_FUNDING_AMOUNT, test_account)
    result = convex.send('map(inc, [1, 2, 3, 4, 5])', test_account)
    assert 'value' in result
    assert(result['value'] == [2, 3, 4, 5, 6])

def test_convex_api_get_balance_no_funds(convex_url):
    convex = ConvexAPI(convex_url)
    account = convex.create_account()
    new_balance = convex.get_balance(account)
    assert(new_balance == 0)

def test_convex_api_get_balance_small_funds(convex_url, test_account):
    convex = ConvexAPI(convex_url)
    account = convex.create_account()
    amount = 100
    request_amount = convex.request_funds(amount, account)
    new_balance = convex.get_balance(account)
    assert(new_balance == amount)

def test_convex_api_get_balance_new_account(convex_url):
    convex = ConvexAPI(convex_url)
    account = convex.create_account()
    amount = TEST_FUNDING_AMOUNT
    request_amount = convex.request_funds(amount, account)
    assert(request_amount == amount)
    new_balance = convex.get_balance(account)
    assert(new_balance == TEST_FUNDING_AMOUNT)

def test_convex_api_call(convex_url):

    deploy_storage = """
(def storage-example
    (deploy
        '(do
            (def stored-data nil)
            (defn get [] stored-data)
            (defn set [x] (def stored-data x))
            (export get set)
        )
    )
)
"""
    convex = ConvexAPI(convex_url)
    account = convex.create_account()
    amount = TEST_FUNDING_AMOUNT
    request_amount = convex.request_funds(amount, account)
    result = convex.send(deploy_storage, account)
    assert(result['value'])
    contract_address = to_address(result['value'])
    test_number = secrets.randbelow(1000)
    call_set_result = convex.send(f'(call storage-example(set {test_number}))', account)
    assert(call_set_result['value'] == test_number)
    call_get_result = convex.query('(call storage-example(get))', account)
    assert(call_get_result['value'] == test_number)

    # now api calls using language scrypt

    convex = ConvexAPI(convex_url, ConvexAPI.LANGUAGE_SCRYPT)
    test_number = secrets.randbelow(1000)
    call_set_result = convex.send(f'call storage_example set({test_number})', account)
    assert(call_set_result['value'] == test_number)

    call_get_result = convex.query('call storage_example get()', account)
    assert(call_get_result['value'] == test_number)

    call_get_result = convex.query(f'call {contract_address} get()', account)
    assert(call_get_result['value'] == test_number)

    with pytest.raises(ConvexRequestError, match='400'):
        call_set_result = convex.send(f'call {contract_address}.set({test_number})', account)

    address = convex.get_address('storage_example', account)
    assert(address == contract_address)

def test_convex_api_transfer(convex_url):
    convex = ConvexAPI(convex_url)
    account_from = convex.create_account()
    account_to = convex.create_account()
    amount = TEST_FUNDING_AMOUNT
    request_amount = convex.request_funds(amount, account_from)
    assert(request_amount == amount)

    transfer_amount = int(amount / 2)
    result = convex.transfer(account_to, transfer_amount, account_from)
    assert(result)
    assert(result == transfer_amount)
    balance_from = convex.get_balance(account_from)
    balance_to = convex.get_balance(account_to)
    assert(balance_to == transfer_amount)

def test_convex_api_query_lisp(convex_url, test_account):
    convex = ConvexAPI(convex_url)
    result = convex.query(f'(address {test_account.address})', test_account)
    assert(result)
    # return value is the address as a checksum
    assert(to_address(result['value']) == test_account.address)

def test_convex_api_query_scrypt(convex_url, test_account):
    convex = ConvexAPI(convex_url, ConvexAPI.LANGUAGE_SCRYPT)
    result = convex.query(f'address({test_account.address})', test_account)
    assert(result)
    # return value is the address as a checksum
    assert(to_address(result['value']) == test_account.address)



"""

    Utils  - address conversions

"""
import re

from cryptography.hazmat.backends.openssl.backend import backend
from cryptography.hazmat.primitives import hashes

from eth_utils import (
    add_0x_prefix,
    is_hexstr,
    remove_0x_prefix,
    to_bytes,
    to_hex
)


def to_address(text):
    if isinstance(text, int):
        return int(text)
    if isinstance(text, str):
        try:
            value = int(re.sub('^#', '', text.strip()))
        except ValueError:
            return None
        return value
    return None


def is_address(text):
    value = to_address(text)
    if isinstance(value, int):
        return value >= 0
    return False


def is_public_key_hex(public_key):
    if is_hexstr(public_key):
        address_base = remove_0x_prefix(public_key)
        if len(address_base) == 64:
            return True
    return False


def is_public_key(public_key):
    if is_public_key_checksum(public_key):
        return True
    if is_public_key_hex(public_key):
        return True
    return False


def to_public_key_checksum(public_key):
    digest = hashes.Hash(hashes.SHA3_256(), backend=backend)
    digest.update(to_bytes(hexstr=public_key))
    public_key_hash = remove_0x_prefix(to_hex(digest.finalize()))
    public_key_clean = remove_0x_prefix(public_key.lower())
    checksum = ''
    hash_index = 0
    for value in public_key_clean:
        if int(public_key_hash[hash_index], 16) > 7:
            checksum += value.upper()
        else:
            checksum += value
        hash_index += 1
        if hash_index >= len(public_key_hash):
            hash_index = 0
    return add_0x_prefix(checksum)


def is_public_key_checksum(public_key):
    return remove_0x_prefix(public_key) and remove_0x_prefix(public_key) == remove_0x_prefix(to_public_key_checksum(public_key))



"""


    Test Convex Breaking

"""

import pytest
import secrets
from eth_utils import add_0x_prefix


from convex_api.account import Account
from convex_api.convex_api import ConvexAPI
from convex_api.exceptions import ConvexAPIError
from convex_api.utils import to_address

def test_convex_recursion(convex, test_account):
    chain_length = 4
    address_list = []
    for index in range(0, chain_length):
        contract = f"""
(def chain-{index}
    (deploy
        '(do
            (def stored-data nil)
            (def chain-address nil)
            (defn get [] (call chain-address (get)))
            (defn set [x] (if chain-address (call chain-address(set x)) (def stored-data x)) )
            (defn set-chain-address [x] (def chain-address x))
            (export get set set-chain-address)
        )
    )
)
"""
        convex.topup_account(test_account)
        result = convex.send(contract, test_account)
        address_list.append(to_address(result['value']))
    for index in range(0, chain_length):
        next_index = index + 1
        if next_index == chain_length:
            next_index = 0
        call_address = address_list[next_index]
        result = convex.send(f'(call chain-{index} (set-chain-address {call_address}))', test_account)
        test_number = secrets.randbelow(1000)
        if index == chain_length - 1:
            with pytest.raises(ConvexAPIError, match='DEPTH'):
                result = convex.send(f'(call chain-{index} (set {test_number}))', test_account)
        else:
            result = convex.send(f'(call chain-0 (set {test_number}))', test_account)
            assert(result)
            assert(result['value'] == test_number)
    with pytest.raises(ConvexAPIError, match='DEPTH'):
        convex.query('(call chain-0 (get))', test_account)

def test_schedule_transfer(convex, test_account, other_account):
    # you can send coins to an actor , if it exports the receive-coin function

    contract = """
(def transfer-for-ever
    (deploy
        '(do
            (defn tx-delay [to-address amount]
                (transfer to-address amount)
                (def call-address *address*)
                (schedule (+ *timestamp* 1000) (call call-address (tx-delay to-address amount)))
            )
            (defn tx-now [to-address amount]
                (transfer to-address amount)
            )
            (defn show-schedule []
                [(get *state* :schedule) *address*]
            )
            (defn receive-coin [sender amount data] (accept amount))
            (export show-schedule tx-delay tx-now, receive-coin)
        )
    )
)
"""
# (call contract-address (tx-to to-address amount))

    convex.topup_account(test_account)
    convex.topup_account(other_account, 8000000)
    result = convex.send(contract, test_account)
    contract_address = to_address(result['value'])
    convex.transfer(contract_address, 800000, other_account)
    convex.topup_account(test_account)
    result = convex.send(f'(call {contract_address} (tx-delay {other_account.address} 1000))', test_account)
    print(result)
    result = convex.send(f'(call {contract_address} (show-schedule))', test_account)
    print(result)



"""
    Test account based functions

"""

from convex_api import (
    Account,
    ConvexAPI
)

TEST_ACCOUNT_NAME = 'test.convex-api'

def test_account_api_create_account(convex_url):

    convex = ConvexAPI(convex_url)
    result = convex.create_account()
    assert(result)


def test_account_api_multi_create_account(convex_url):
    convex = ConvexAPI(convex_url)
    account = Account.create()
    account_1 = convex.create_account(account=account)
    assert(account_1)
    account_2 = convex.create_account(account=account)
    assert(account_2)

    assert(account.public_key == account_1.public_key)
    assert(account.public_key == account_2.public_key)
    assert(not account.is_address)
    assert(account_1.address != account_2.address)


def test_account_name(convex_url, test_account_info):
    convex = ConvexAPI(convex_url)
    import_account = Account.import_from_bytes(test_account_info['private_bytes'])
    if convex.resolve_account_name(TEST_ACCOUNT_NAME):
        account = convex.load_account(TEST_ACCOUNT_NAME, import_account)
    else:
        account = convex.create_account(account=import_account)
        convex.topup_account(account)
        account = convex.register_account_name(TEST_ACCOUNT_NAME, account)
    assert(account.address)
    assert(account.name)
    assert(account.name == TEST_ACCOUNT_NAME)
    assert(convex.resolve_account_name(TEST_ACCOUNT_NAME) == account.address)


def test_account_setup_account(convex_url, test_account_info):
    convex = ConvexAPI(convex_url)
    import_account = Account.import_from_bytes(test_account_info['private_bytes'])
    account = convex.setup_account(TEST_ACCOUNT_NAME, import_account)
    assert(account.address)
    assert(account.name)
    assert(account.name == TEST_ACCOUNT_NAME)
    assert(convex.resolve_account_name(TEST_ACCOUNT_NAME) == account.address)



"""

    Convex API

"""

import json
import logging
import re
import secrets
import time

from urllib.parse import urljoin

import requests
from eth_utils import remove_0x_prefix

from convex_api.account import Account

from convex_api.exceptions import (
    ConvexAPIError,
    ConvexRequestError
)
from convex_api.registry import Registry
from convex_api.utils import (
    is_address,
    to_address
)

# min amount to do a topup account
TOPUP_ACCOUNT_MIN_BALANCE = 10000000

logger = logging.getLogger(__name__)


class ConvexAPI:

    LANGUAGE_LISP = 'convex-lisp'
    LANGUAGE_SCRYPT = 'convex-scrypt'
    LANGUAGE_ALLOWED = [LANGUAGE_LISP, LANGUAGE_SCRYPT]

    def __init__(self, url, language=LANGUAGE_LISP):
        self._url = url

        if language not in ConvexAPI.LANGUAGE_ALLOWED:
            raise ValueError(f'Invalid language: {language}')
        self._language = language
        self._registry = Registry(self)

    def create_account(self, account=None, sequence_retry_count=20):
        """

        Create a new account address on the convex network.

        :param `Account` account: :class:`.Account` object that you whish to use as the signing account.
            The :class:`.Account` object contains the public/private keys to access and submit commands
            on the convex network.

            If no object given, then this method will automatically create a new :class:`.Account` object.
        :type account: Account, optional


        :param sequence_retry_count: Number of retries to create the account. If too many clients are trying to
            create accounts on the same node, then we will get sequence errors.
        :type sequence_retry_count: int, optional

        :returns: A new :class:`.Account` object, or copy of the :class:`.Account` object with a new `address` property value set


        .. code-block:: python

            >>> from convex_api import ConvexAPI
            >>> convex_api = ConvexAPI('https://convex.world')
            >>> # Create a new account with new public/priavte keys and address
            >>> account = convex_api.create_account()
            >>> print(account.address)
            >>> 42

            >>> #create a new account address, but use the same keys as `account`
            >>> new_account_address = convex_api.create_account(account=account)
            >>> print(new_account.address)
            >>> 43


        """

        if account and not isinstance(account, Account):
            raise TypeError(f'account value {account} must be a type convex_api.Account')

        # if account, then copy it
        if account:
            new_account = Account.import_from_account(account)
        else:
            # create a new account with new keys
            new_account = Account.create()

        create_account_url = urljoin(self._url, '/api/v1/createAccount')
        account_data = {
            'accountKey': remove_0x_prefix(new_account.public_key),
        }

        logger.debug(f'create_account {create_account_url} {account_data}')

        max_sleep_time_seconds = 1
        while sequence_retry_count >= 0:
            response = requests.post(create_account_url, data=json.dumps(account_data))
            if response.status_code == 200:
                result = response.json()
                break
            elif response.status_code == 400:
                if not re.search(':SEQUENCE ', response.text):
                    raise ConvexRequestError('create_account', response.status_code, response.text)

                if sequence_retry_count == 0:
                    raise ConvexRequestError('create_account', response.status_code, response.text)
                sequence_retry_count -= 1
                # now sleep < 1 second for at least 1 millisecond
                sleep_time = secrets.randbelow(round(max_sleep_time_seconds * 1000)) / 1000
                time.sleep(sleep_time + 1)
            else:
                raise ConvexRequestError('create_account', response.status_code, response.text)
        logger.debug(f'create_account result {result}')
        new_account.address = to_address(result['address'])

        return new_account

    def load_account(self, name, account):
        """

        Load an account using the account name. If successfull return the :class:`.Account` object with the address set.

        This is a Query operation, so no convex tokens are used in loading the account.

        :param str name: name of the account to load
        :param Account account: :class:`.Account` object to import

        :results: :class:`.Account` object with the address and name set, if not found then return None


        .. code-block:: python

            >>> # Create a new account with new public/priavte keys and address
            >>> import_account = Account.import_from_file('my_account.pem', 'secret')
            >>> account = convex.load_account('my_account, import_account)
            >>> print(account.name)
            my_account
            >>> print(account.address)
            930

        """
        address = self.resolve_account_name(name)
        if address:
            new_account = Account.import_from_account(account, address=address, name=name)
            return new_account

    def setup_account(self, name, import_account):
        """

        Convenience method to create or load an account based on the account name.
        If the account name cannot be found then account will be created and account name registered,
        if the name is found, then account and it's address with that name will be loaded.

        :param str name: name of the account to create or load
        :param Account account: :class:`.Account` object to import

        :results: :class:`.Account` object with the address and name set, if not found then return None

        **Note** This method calls the :meth:`.topup_account` method to get enougth funds to register an account name.


        .. code-block:: python

            >>> import_account = Account.import_from_file('my_account.pem', 'secret')
            >>> # create or load the account named 'my_account'
            >>> account = convex.setup_account('my_account', import_account)
            >>> print(account.name)
            my_account

        """
        if self.resolve_account_name(name):
            account = self.load_account(name, import_account)
        else:
            account = self.create_account(account=import_account)
            self.topup_account(account)
            self.register_account_name(name, account)
        self.topup_account(account)
        return account

    def register_account_name(self, name, account):
        """

        Register an account address with an account name.

        This call will submit to the CNS (Convex Name Service), a name in the format
        "`account.<your_name>`". You need to have some convex balance in your account, and
        a valid account address.

        :param str name: name of the account to register
        :param Account account: :class:`.Account` object to register the account name


        >>> # create a new account
        >>> account = convex.create_account()
        >>> # add some convex tokens to the account
        >>> convex.topup_account(account)
        10000000
        >>> account = convex.register_account('my_new_account', account)
        >>> print(account.name)
        my_new_account

        """
        if account.address:
            self._registry.register(f'account.{name}', account.address, account)
            return Account.import_from_account(account, address=account.address, name=name)

    def send(self, transaction, account, language=None, sequence_retry_count=20):
        """
        Send transaction code to the block chain node.

        :param str transaction: The transaction as a string to send

        :param Account account: The account that needs to sign the message to send

        :param language: Language to use for this transaction. Defaults to LANGUAGE_LISP.
        :type language: str, optional

        :param sequence_retry_count: Number of retries to do if a SEQUENCE error occurs.
            When sending multiple send requsts on the same account, you can get SEQUENCE errors,
            This send method will automatically retry again
        :type sequence_retry_count: int, optional

        :returns: The dict returned from the result of the sent transaction.


        .. code-block:: python

            >>> from convex_api import ConvexAPI
            >>> convex_api = ConvexAPI('https://convex.world')

            >>> # Create a new account with new public/priavte keys and address
            >>> account = convex_api.create_account()

            >>> # request some funds to do stuff
            >>> print(convex_api.request_funds(100000, account))
            100000

            >>> # submit a transaction using the new account
            >>> print(convex_api.send('(map inc [ 1 2 3 4])', account))
            {'value': [2, 3, 4, 5]}


        """
        if not transaction:
            raise ValueError('You need to provide a valid transaction')
        if not isinstance(transaction, str):
            raise TypeError('The transaction must be a type str')

        result = None
        max_sleep_time_seconds = 1
        while sequence_retry_count >= 0:
            try:
                hash_data = self._transaction_prepare(account.address, transaction, language=language)
                signed_data = account.sign(hash_data['hash'])
                result = self._transaction_submit(account.address, account.public_key_checksum, hash_data['hash'], signed_data)
            except ConvexAPIError as error:
                if error.code == 'SEQUENCE':
                    if sequence_retry_count == 0:
                        raise
                    sequence_retry_count -= 1
                    # now sleep < 1 second for at least 1 millisecond
                    sleep_time = secrets.randbelow(round(max_sleep_time_seconds * 1000)) / 1000
                    time.sleep(sleep_time + 1)
                else:
                    raise
            else:
                break
        return result

    def query(self, transaction, address_account, language=None):
        """

        Run a query transaction on the block chain. Since this does not change the network state, and
        the account does not need to sign the transaction. No funds will be used when executing
        this query. For this reason you can just pass the account address, or if you want to the :class:`.Account` object.

        :param str transaction: Transaction to execute. This can only be a read only transaction.

        :param address_account: :class:`.Account` object or int address of an account to use for running this query.
        :type address_account: Account, int, str

        :param language: The type of language to use, if not provided the default language set will be used.
        :type language: str, optional

        :returns: Return the resultant query transaction


        .. code-block:: python

            >>> # Create a new account with new public/priavte keys and address
            >>> account = convex_api.create_account()

            >>> # submit a query transaction using the account address

            >>> print(convex_api.query(f'(balance {account.address})', account.address))
            {'value': 0}

            >>> # request some funds to do stuff
            >>> print(convex_api.request_funds(100000, account))
            100000
            >>> print(convex_api.query(f'(balance {account.address})', account.address))
            {'value': 100000}

        """
        if is_address(address_account):
            address = to_address(address_account)
        else:
            address = address_account.address

        return self._transaction_query(address, transaction, language)

    def request_funds(self, amount, account):
        """

        Request funds for an account from the block chain faucet.

        :param number amount: The amount of funds to request

        :param Account account: The :class:`.Account` object to receive funds too

        :returns: The amount transfered to the account


        .. code-block:: python

            >>> # Create a new account with new public/priavte keys and address
            >>> account = convex_api.create_account()
            >>> # request some funds to do stuff
            >>> print(convex_api.request_funds(100000, account))
            100000

        """
        faucet_url = urljoin(self._url, '/api/v1/faucet')
        faucet_data = {
            'address': account.address,
            'amount': amount
        }
        logger.debug(f'request_funds {faucet_url} {faucet_data}')
        response = requests.post(faucet_url, data=json.dumps(faucet_data))
        if response.status_code != 200:
            raise ConvexRequestError('request_funds', response.status_code, response.text)
        result = response.json()
        logger.debug(f'request_funds result {result}')
        if result['address'] != account.address:
            raise ValueError(f'request_funds: returned account is not correct {result["address"]}')
        return result['amount']

    def topup_account(self, account, min_balance=TOPUP_ACCOUNT_MIN_BALANCE, retry_count=8):
        """

        Topup an account from the block chain faucet, so that the balance of the account is above or equal to
        the `min_balanace`.

        :param Account account: The :class:`.Account` object to receive funds for

        :param min_balance: Minimum account balance that will allowed before a topup occurs
        :type min_balance: number, optional

        :param retry_count: The number of times the faucet will be called to get above or equal to the  `min_balance`
        :type retry_count: number, optional

        :returns: The amount transfered to the account

        .. code-block:: python

            >>> # Create a new account with new public/priavte keys and address
            >>> account = convex_api.create_account()
            >>> # request some funds to do stuff
            >>> print(convex_api.topup_account(account, 100000))
            100000
            >>> # try again, but only return 0 topup amount credited
            >>> print(convex_api.topup_account(account, 100000))
            0

        """

        request_amount = min(TOPUP_ACCOUNT_MIN_BALANCE, min_balance)
        retry_count = min(5, retry_count)
        transfer_amount = 0
        while min_balance > self.get_balance(account) and retry_count > 0:
            transfer_amount += self.request_funds(request_amount, account)
            retry_count -= 1
        return transfer_amount

    def get_address(self, function_name, address_account):
        """

        Query the network for a contract ( function ) address. The contract must have been deployed
        by the account address provided. If not then no address will be returned

        :param str function_name: Name of the contract/function

        :param address_account: :class:`.Account` object or str address of an account to use for running this query.
        :type address_account: Account, int, str


        :returns: Returns address of the contract

        .. code-block:: python

            >>> # Create a new account with new public/priavte keys and address
            >>> account = convex_api.create_account()
            >>> # find the address of a contract
            >>> print(convex_api.get_address('my_contract', account))

        """

        line = f'(address {function_name})'
        if self._language == ConvexAPI.LANGUAGE_SCRYPT:
            line = f'address({function_name})'
        result = self.query(line, address_account)
        if result and 'value' in result:
            return to_address(result['value'])

    def get_balance(self, address_account, account_from=None):
        """

        Get a balance of an account.

        At the moment the account needs to have a balance to get the balance of it's account or any
        other account. Event though this is using a query request.

        :param address_account: Address or :class:`.Account` object to get the funds for.
        :type address_account: Account, int, str

        :param account_from: Optional :class:`.Account` object or account address to make the request.
        :type account_from: Account, int, str, optional


        :returns: Return the current balance of the address or account `address_account`

        .. code-block:: python

            >>> # Create a new account with new public/priavte keys and address
            >>> account = convex_api.create_account()
            >>> # get the balance of the contract
            >>> print(convex_api.get_balance(account))
            0
            >>> print(convex_api.request_funds(100000, account))
            100000
            >>> print(convex_api.get_balance(account))
            100000

        """
        value = 0
        if is_address(address_account):
            address = to_address(address_account)
        else:
            address = address_account.address

        address_from = address
        if account_from:
            if is_address(account_from):
                address_from = to_address(account_from)
            else:
                address_from = account_from.address
        line = f'(balance {address})'
        if self._language == ConvexAPI.LANGUAGE_SCRYPT:
            line = f'balance({address})'
        try:

            result = self._transaction_query(address_from, line)
        except ConvexAPIError as error:
            if error.code != 'NOBODY':
                raise
        else:
            value = result['value']
        return value

    def transfer(self, to_address_account, amount, account):
        """

        Transfer funds from on account to another.

        :param to_address_account: Address or :class:`.Account` object to send the funds too
        :type to_address_account: Account, int, str

        :param number amount: Amount to transfer

        :param Account account: :class:`.Account` object to send the funds from

        :returns: The transfer record sent back after the transfer has been made

        .. code-block:: python

            >>> # Create a new account with new public/priavte keys and address
            >>> account = convex_api.create_account()
            >>> print(convex_api.request_funds(10000000, account))
            10000000
            >>> print(convex_api.get_balance(account))
            10000000

            >>> my_account = convex_api.create_account()
            >>> # transfer some funds to my_account
            >>> print(convex_api.transfer(my_account, 100, account))
            100
            >>> print(convex_api.get_balance(my_account))
            100
            >>> print(convex_api.get_balance(account))
            9998520

        """
        if is_address(to_address_account):
            transfer_to_address = to_address(to_address_account)
        else:
            transfer_to_address = to_address_account.address
        if not to_address:
            raise ValueError(f'You must provide a valid to account/address ({transfer_to_address}) to transfer funds too')

        line = f'(transfer {transfer_to_address} {amount})'
        if self._language == ConvexAPI.LANGUAGE_SCRYPT:
            line = f'transfer({transfer_to_address}, {amount})'

        result = self.send(line, account)
        if result and 'value' in result:
            return result['value']
        return 0

    def get_account_info(self, address_account):
        """

        Get account information. This will only work with an account that has a balance or has had some transactions
        processed on the convex network. New accounts with no transfer or transactions will raise:

            ConvexRequestError(404, 'The Account for this Address does not exist.') error

        The returned information is dictionary of account information.

        :param address_account: :class:`.Account` object or address of an account to get current information on.
        :type address_account: Account, int, str

        :returns: dict of information, such as

        .. code-block:: python

            >>> # Create a new account with new public/priavte keys and address
            >>> account = convex_api.create_account()
            >>> # get the balance of the contract
            >>> print(convex_api.get_account_info(account))

            {'environment': {}, 'address': 1178, 'memorySize': 0, 'balance': 0,
            'isLibrary': False, 'isActor': False, 'allowance': 0,
            'sequence': 0, 'type': 'user'}




        """
        if is_address(address_account):
            address = to_address(address_account)
        else:
            address = address_account.address

        account_url = urljoin(self._url, f'/api/v1/accounts/{address}')
        logger.debug(f'get_account_info {account_url}')

        response = requests.get(account_url)
        if response.status_code != 200:
            raise ConvexRequestError('get_account_info', response.status_code, response.text)

        result = response.json()
        logger.debug(f'get_account_info repsonse {result}')
        return result

    def resolve_account_name(self, name):
        return self._registry.resolve_address(f'account.{name}')

    def _transaction_prepare(self, address, transaction, language=None, sequence_number=None):
        """

        """
        if language is None:
            language = self._language
        prepare_url = urljoin(self._url, '/api/v1/transaction/prepare')
        data = {
            'address': address,
            'lang': language,
            'source': transaction,
        }
        if sequence_number:
            data['sequence'] = sequence_number
        logger.debug(f'_transaction_prepare {prepare_url} {data}')
        response = requests.post(prepare_url, data=json.dumps(data))
        if response.status_code != 200:
            raise ConvexRequestError('_transaction_prepare', response.status_code, response.text)

        result = response.json()
        logger.debug(f'_transaction_prepare repsonse {result}')
        if 'errorCode' in result:
            raise ConvexAPIError('_transaction_prepare', result['errorCode'], result['value'])

        return result

    def _transaction_submit(self, address, public_key, hash_data, signed_data):
        """

        """
        submit_url = urljoin(self._url, '/api/v1/transaction/submit')
        data = {
            'address': to_address(address),
            'accountKey': remove_0x_prefix(public_key),
            'hash': hash_data,
            'sig': remove_0x_prefix(signed_data)
        }
        logger.debug(f'_transaction_submit {submit_url} {data}')
        response = requests.post(submit_url, data=json.dumps(data))
        if response.status_code != 200:
            raise ConvexRequestError('_transaction_submit', response.status_code, response.text)

        result = response.json()
        logger.debug(f'_transaction_submit response {result}')
        if 'errorCode' in result:
            raise ConvexAPIError('_transaction_submit', result['errorCode'], result['value'])
        return result

    def _transaction_query(self, address, transaction, language=None):
        """

        """
        if language is None:
            language = self._language

        prepare_url = urljoin(self._url, '/api/v1/query')
        data = {
            'address': address,
            'lang': language,
            'source': transaction,
        }
        logger.debug(f'_transaction_query {prepare_url} {data}')
        response = requests.post(prepare_url, data=json.dumps(data))
        if response.status_code != 200:
            raise ConvexRequestError('_transaction_query', response.status_code, response.text)

        result = response.json()
        logger.debug(f'_transaction_query repsonse {result}')
        if 'errorCode' in result:
            raise ConvexAPIError('_transaction_query', result['errorCode'], result['value'])
        return result

    @property
    def language(self):
        """

        Returns the default language to use when calling the API commands

        """
        return self._language



"""

    Test Account class

"""
import os
import secrets


from convex_api.account import Account
from eth_utils import (
    remove_0x_prefix,
    to_bytes,
    to_checksum_address,
    to_hex
)


SIGN_HASH_TEXT = '5bb1ce718241bfec110552b86bb7cccf0d95b8a5f462fbf6dff7c48543622ba5'
SIGN_TEXT = '0x7eceffab47295be3891ea745838a99102bfaf525ec43632366c7ec3f54db4822b5d581573aecde94c420554f963baebbf412e4304ad8636886ddfa7b1049f70e'
def test_account_create_new():
    account = Account.create()
    assert(account)
    assert(account.public_key)


def test_account_create_from_bytes(test_account_info):
    account = Account.import_from_bytes(test_account_info['private_bytes'])
    assert(account)
    assert(account.public_key == test_account_info['public_key'])

def test_account_address_bytes(test_account_info):
    account = Account.import_from_bytes(test_account_info['private_bytes'])
    assert(account)
    assert(account.public_key_bytes == to_bytes(hexstr=test_account_info['public_key']))

def test_account_address_api(test_account_info):
    account = Account.import_from_bytes(test_account_info['private_bytes'])
    assert(account)
    assert(account.public_key_api == remove_0x_prefix(test_account_info['public_key']))

def test_account_address_checksum(test_account_info):
    account = Account.import_from_bytes(test_account_info['private_bytes'])
    assert(account)
    assert(account.public_key_checksum.lower() == test_account_info['public_key'])

def test_account_sign(test_account_info):
    hash_text = SIGN_HASH_TEXT
    account = Account.import_from_bytes(test_account_info['private_bytes'])
    sign_data = account.sign(hash_text)
    assert(sign_data == SIGN_TEXT)


def test_account_import_export_to_text(test_account):
    password = 'secret'
    text = test_account.export_to_text(password)
    import_account = Account.import_from_text(text, password)
    assert(import_account)
    assert(import_account.public_key == test_account.public_key)


def test_account_import_export_to_file(test_account):
    filename = '/tmp/private_key.pem'
    password = 'secret'
    if os.path.exists(filename):
        os.remove(filename)

    text = test_account.export_to_file(filename, password)
    assert(os.path.exists(filename))
    import_account = Account.import_from_file(filename, password)
    assert(import_account)
    assert(import_account.public_key == test_account.public_key)
    os.remove(filename)

def test_account_export_to_mnemonic(test_account):
    words = test_account.export_to_mnemonic
    assert(words)
    new_account = Account.import_from_mnemonic(words)
    assert(new_account)
    assert(test_account.public_key == new_account.public_key)
    assert(test_account.export_to_mnemonic == new_account.export_to_mnemonic)



"""

    Test Convex multi thread

"""
import pytest
import secrets
from multiprocessing import (
    Process,
    Value
)

from convex_api.convex_api import ConvexAPI
from convex_api.exceptions import ConvexAPIError
from convex_api.utils import to_address

TEST_FUNDING_AMOUNT = 8000000

def process_on_convex(convex, test_account, result_value):
    values = []
    inc_values = []
    is_sent = False
    for counter in range(0, 4):
        for index in range(secrets.randbelow(10) + 1):
            value = secrets.randbelow(1000)
            values.append(str(value))
            inc_values.append(value + 1)
            value_text = " ".join(values)
        result = convex.send(f'(map inc [{value_text}])', test_account, sequence_retry_count=100)
        assert(result)
        assert('value' in result)
        assert(result['value'] == inc_values)
    result_value.value = 1


def test_convex_api_multi_thread_send(convex_url, test_account):

    process_count = 4
    convex = ConvexAPI(convex_url)
    convex.topup_account(test_account)
    process_items = {}
    for index in range(process_count):
        result_value = Value('i', 0)
        proc = Process(target=process_on_convex, args=(convex, test_account, result_value))
        process_items[index] = {
            'process': proc,
            'result_value': result_value
        }
        proc.start()

    for index, process_item in process_items.items():
        process_item['process'].join()
        assert(process_item['result_value'].value == 1)


def process_convex_account_creation(convex, result_value):
    account = convex.create_account()
    assert(account)
    assert(account.address)
    result_value.value = 1


def test_convex_api_multi_thread_account_creation(convex_url):
    process_count = 20
    convex = ConvexAPI(convex_url)
    process_items = {}
    for index in range(process_count):
        result_value = Value('i', 0)
        proc = Process(target=process_convex_account_creation, args=(convex, result_value))
        process_items[index] = {
            'process': proc,
            'result_value': result_value
        }
        proc.start()

    for index, process_item in process_items.items():
        process_item['process'].join()
        assert(process_item['result_value'].value == 1)


def process_convex_depoly(convex, result_value):
    deploy_storage = """
(def storage-example
    (deploy
        '(do
            (def stored-data nil)
            (defn get [] stored-data)
            (defn set [x] (def stored-data x))
            (export get set)
        )
    )
)
"""
    account = convex.create_account()
    for index in range(0, 10):
        convex.topup_account(account)
        try:
            result = convex.send(deploy_storage, account)
        except ConvexAPIError as e:
            balance = convex.get_balance(account)
            print('*' * 132)
            print('failed send', e, balance)
            print('*' * 132)
            result_value.value = balance
            return
        assert(result)
        assert(result['value'])
        contract_address = to_address(result['value'])
        assert(contract_address)
    result_value.value = 1


def test_convex_api_multi_thread_deploy(convex_url):
    process_count = 10
    convex = ConvexAPI(convex_url)
    # account = convex.create_account()
    # request_amount = convex.request_funds(TEST_FUNDING_AMOUNT, account)
    process_items = {}
    for index in range(process_count):
        result_value = Value('i', 0)
        proc = Process(target=process_convex_depoly, args=(convex, result_value))
        process_items[index] = {
            'process': proc,
            'result_value': result_value
        }
        proc.start()

    for index, process_item in process_items.items():
        process_item['process'].join()
        assert(process_item['result_value'].value == 1)




#!/usr/bin/env python3

"""

    Script to provide convex wallet functionality

"""

import argparse
import json
import logging
import secrets

from convex_api import Account as ConvexAccount
from convex_api import ConvexAPI
from convex_api.utils import is_address


DEFAULT_URL = 'https://convex.world'

COMMAND_HELP_TEXT = '''

create                      Create a new account using the provided --password. If no password auto generate one.
new                         Same as 'create' account command.
info [address]              Get information about an account, you can pass the account address, or the options <keywords> or <keyfile>/<password> of the account.

'''         # noqa: E501

logger = logging.getLogger('convex_wallet')


def load_account(args):
    account = None
    if args.keyfile and args.password:
        account = ConvexAccount.import_from_file(args.keyfile, args.password)
    elif args.keywords:
        account = ConvexAccount.import_from_mnemonic(args.keywords)
    return account


def main():

    parser = argparse.ArgumentParser(
        description='Convex Wallet',
        formatter_class=argparse.RawTextHelpFormatter
    )

    parser.add_argument(
        '--auto-topup',
        action='store_true',
        help='Auto topup account with sufficient funds. This only works for development networks. Default: False',
    )

    parser.add_argument(
        '-d',
        '--debug',
        action='store_true',
        help='Debug mode on or off. Default: False',
    )

    parser.add_argument(
        '-k',
        '--keyfile',
        nargs='?',
        help='account private key encrypted with password saved in a file'
    )

    parser.add_argument(
        '-p',
        '--password',
        nargs='?',
        help='password to access the private key enrcypted in a keyfile'
    )

    parser.add_argument(
        '-w',
        '--keywords',
        nargs='?',
        help='account private key as words'
    )

    parser.add_argument(
        '-n',
        '--name',
        nargs='?',
        help='account name to register'
    )

    parser.add_argument(
        '-u',
        '--url',
        default=DEFAULT_URL,
        help=f'URL of the network node. Default: {DEFAULT_URL}',
    )

    parser.add_argument(
        'command',
        help=f'Wallet commands are as follows: \r\n{COMMAND_HELP_TEXT}'
    )

    parser.add_argument(
        'command_args',
        nargs='*',
    )

    args = parser.parse_args()
    convex = ConvexAPI(args.url)
    if not convex:
        print(f'Cannot connect to the convex network at {args.url}')
        return

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
        logging.getLogger('urllib3').setLevel(logging.INFO)

    if args.command == 'create' or args.command == 'new':
        import_account = load_account(args)
        account = convex.create_account(account=import_account)

        if args.auto_topup:
            logger.debug('auto topup of account balance')
            convex.topup_account(account)

        if args.name:
            convex.topup_account(account)
            account = convex.register_account_name(args.name, account)

        if args.password:
            password = args.password
        else:
            password = secrets.token_hex(32)

        values = {
            'password': password,
            'address': account.address,
            'keyfile': account.export_to_text(password),
            'keywords': account.export_to_mnemonic,
            'balance': convex.get_balance(account)
        }
        if account.name:
            values['name'] = account.name

        print(json.dumps(values, sort_keys=True, indent=4))
    elif args.command == 'info':
        address = None
        if len(args.command_args) > 0:
            if is_address(args.command_args[0]):
                address = args.command_args[0]
            else:
                name = args.command_args[0]
                address = convex.resolve_account_name(name)
        if not address:
            print('cannot find account address using name or address provided')
            return
        values = convex.get_account_info(address)
        print(json.dumps(values, sort_keys=True, indent=4))


if __name__ == "__main__":
    main()
